// -*- Mode: c++ -*-
// $Id: dualmesh.cpp,v 1.12 2000/08/08 21:28:16 jianbo Exp $
// $Source: /tools/cvs-repos/illustrate/src/dualmesh.cpp,v $

#if defined(_WIN32) || defined(__sgi)
#include <float.h>
#endif
#include "cvec4t.h"
#include "options.h"
#include "dualmesh.h"
#include "silhouette.h"

vector<DualEdge*> DualEdge::_checkedEdges;

// boundary edges
vector<BoundaryEdge> OcTree::boundary;
const int OcTree::MAX_EDGE_NUMBER = 10;

// eight surfaces of the 4D cube
enum { PX = 0, PY = 1, PZ = 2, PW = 3, NX = 4, NY = 5, NZ = 6, NW = 7 };

// DualEdge constructor
// assume pos1 and pos2 will project onto the tp-th surface
DualEdge::DualEdge(Tri *t1, Tri *t2, EnoType e1, EnoType e2, bool fac,
		   int tp, const cvec4d& pos1, const cvec4d& pos2)
  : _facing(fac), _checked(false) {
  _tri[0] = t1, _tri[1] = t2, _eno[0] = e1, _eno[1] = e2;

  double d1, d2;
  switch (tp) {
  case PX:
  case NX:
    d1 = ::max(DBL_EPSILON, fabs(pos1.x()));
    d2 = ::max(DBL_EPSILON, fabs(pos2.x()));
    _dv[0] = cvec3d(pos1.y()/d1, pos1.z()/d1, pos1.w()/d1);
    _dv[1] = cvec3d(pos2.y()/d2, pos2.z()/d2, pos2.w()/d2);
    break;
  case PY:
  case NY:
    d1 = ::max(DBL_EPSILON, fabs(pos1.y()));
    d2 = ::max(DBL_EPSILON, fabs(pos2.y()));
    _dv[0] = cvec3d(pos1.z()/d1, pos1.w()/d1, pos1.x()/d1);
    _dv[1] = cvec3d(pos2.z()/d2, pos2.w()/d2, pos2.x()/d2);
    break;
  case PZ:
  case NZ:
    d1 = ::max(DBL_EPSILON, fabs(pos1.z()));
    d2 = ::max(DBL_EPSILON, fabs(pos2.z()));
    _dv[0] = cvec3d(pos1.w()/d1, pos1.x()/d1, pos1.y()/d1);
    _dv[1] = cvec3d(pos2.w()/d2, pos2.x()/d2, pos2.y()/d2);
    break;
  case PW:
  case NW:
    d1 = ::max(DBL_EPSILON, fabs(pos1.w()));
    d2 = ::max(DBL_EPSILON, fabs(pos2.w()));
    _dv[0] = cvec3d(pos1.x()/d1, pos1.y()/d1, pos1.z()/d1);
    _dv[1] = cvec3d(pos2.x()/d2, pos2.y()/d2, pos2.z()/d2);
  }
}

// intersect a box
bool DualEdge::touch(const cvec3d& minp, const cvec3d& maxp) const {
  int m = 4;
  cvec3d step = (_dv[1] - _dv[0]) / m;
  cvec3d A = _dv[0], B = _dv[0] + step;
  for (int i = 0; i < m; i++) {
    cvec3d bmin = A.min(B), bmax = A.max(B);
    if (bmin.x() <= maxp.x() && minp.x() <= bmax.x() &&
	bmin.y() <= maxp.y() && minp.y() <= bmax.y() &&
	bmin.z() <= maxp.z() && minp.z() <= bmax.z()) return true;
    A = B;
    B += step;
  }
  return false;
}

void DualEdge::project(OcTree *dmesh, Tri *tri, EnoType eno) {
  // neighbor sharing the edge
  EnoType neno;
  Tri *neighbor = tri->neighbor(eno, neno);
  // insure every internal edge is projected exactly once
  if ((neighbor != 0) && (neighbor <= tri)) return;

  // boundary edges
  if (neighbor == 0) {
    OcTree::boundary.push_back(BoundaryEdge(tri, eno));
    return;
  }

  // internal edges are projected

  // facing of the edge
  double ff = tri->normal().dot
    (neighbor->headVert(neighbor->nextEno(neno))->limitPos() -
     tri->headVert(tri->nextEno(eno))->limitPos());

#ifndef NDEBUG
  if (fabs(ff) < Options::EPSILON)
    cerr << "WARNING : potential facing precision problem " << ff << endl;
#endif

  bool fac = (ff > 0);

  cvec3d trn = tri->normal();
  cvec4d n1( trn.x(), trn.y(), trn.z(), - trn.dot(tri->vert(0)->limitPos()) );
  cvec4d N1 = n1.abs();

  trn = neighbor->normal();
  cvec4d n2( trn.x(), trn.y(), trn.z(), - trn.dot(neighbor->vert(0)->limitPos()) );
  cvec4d N2 = n2.abs();

  cvec4d n = n2 - n1;

  double t[14] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1};

  // if the edge should be projected on two surfaces, it will intersect
  // the plane passing orig and the intersection line of the two surfaces
  // these are the 12 possible intersections
  if (n.x() - n.y() != 0) t[1] = - (n1.x() - n1.y()) / (n.x() - n.y());
  if (n.x() - n.z() != 0) t[2] = - (n1.x() - n1.z()) / (n.x() - n.z());
  if (n.x() - n.w() != 0) t[3] = - (n1.x() - n1.w()) / (n.x() - n.w());
  if (n.x() + n.y() != 0) t[4] = - (n1.x() + n1.y()) / (n.x() + n.y());
  if (n.x() + n.z() != 0) t[5] = - (n1.x() + n1.z()) / (n.x() + n.z());
  if (n.x() + n.w() != 0) t[6] = - (n1.x() + n1.w()) / (n.x() + n.w());
  if (n.y() - n.z() != 0) t[7] = - (n1.y() - n1.z()) / (n.y() - n.z());
  if (n.y() - n.w() != 0) t[8] = - (n1.y() - n1.w()) / (n.y() - n.w());
  if (n.y() + n.z() != 0) t[9] = - (n1.y() + n1.z()) / (n.y() + n.z());
  if (n.y() + n.w() != 0)t[10] = - (n1.y() + n1.w()) / (n.y() + n.w());
  if (n.z() - n.w() != 0)t[11] = - (n1.z() - n1.w()) / (n.z() - n.w());
  if (n.z() + n.w() != 0)t[12] = - (n1.z() + n1.w()) / (n.z() + n.w());

  int p1 = 0, p2;
  while (p1 < 13) {
    p2 = 13;
    for (int i=1; i<13; i++) {
      if (t[i] > t[p1] && t[i] < t[p2])	p2 = i;
    }

    // the part that is projected on one of the surfaces
    cvec4d pos1 = n1 + t[p1]*n, pos2 = n1 + t[p2]*n;
    cvec4d mid = 0.5*(pos1+pos2);
    cvec4d MID = mid.abs();

    p1 = p2;

    // the surface projected on
    int tp;
    if (MID.x() >= MID.y() && MID.x() >= MID.z() && MID.x() >= MID.w())
      tp = mid.x() > 0 ? PX : NX;
    else if (MID.y() >= MID.z() && MID.y() >= MID.w())
      tp = mid.y() > 0 ? PY : NY;
    else if (MID.z() >= MID.w())
      tp = mid.z() > 0 ? PZ : NZ;
    else
      tp = mid.w() > 0 ? PW : NW;

    // project this part on the surface
    DualEdge *dedge = new DualEdge(tri, neighbor, eno, neno, fac,
				   tp, pos1, pos2);
    dmesh[tp].insert(dedge);
  }
}

OcTree::~OcTree() {
  for (int i=0; i<8; i++) {
    if (_child[i]) delete _child[i];
    _child[i] = 0;
  }
  // delete dual edges
  if (_depth == 0) {
    int sz = _edges.size();
    for (int j=0; j<sz; j++)
      delete _edges[j];
  }
  _edges.clear();
}

void OcTree::split() {
  int esz = _edges.size();

  if (esz <= MAX_EDGE_NUMBER || 
      (esz <= 2*MAX_EDGE_NUMBER && _depth > 3) ||
      (esz <= 3*MAX_EDGE_NUMBER && _depth > 4) ||
      _depth > 6 ) {
    _isLeaf = true;
    return;
  }

  _isLeaf = false;

  double xx[] = {_min.x(), 0.5*(_min.x()+_max.x()), _max.x()};
  double yy[] = {_min.y(), 0.5*(_min.y()+_max.y()), _max.y()};
  double zz[] = {_min.z(), 0.5*(_min.z()+_max.z()), _max.z()};
  for (int i = 0; i < 2; i++)
    for (int j = 0; j < 2; j++)
      for (int k = 0; k < 2; k++) {
	int m = 4*i + 2*j + k;
	_child[m] = new OcTree( cvec3d(xx[i], yy[j], zz[k]),
				cvec3d(xx[i+1], yy[j+1], zz[k+1]),
				_depth+1 );
	int esz = _edges.size();
	for (int t = 0; t < esz; t++)
	  if (_edges[t]->touch(_child[m]->minp(), _child[m]->maxp()))
	    _child[m]->insert(_edges[t]);
	if (_child[m]->size() == 0) {
	  delete _child[m];
	  _child[m] = 0;
	} else _child[m]->split();
      }
  if (_depth > 0) _edges.clear();
}

// intersect a plane
bool OcTree::touchPlane(const cvec3d& n, double d) {
  bool sa = n.x()>=0, sb = n.y()>=0, sc = n.z()>=0;
  double p1x = _min.x(), p1y, p1z, p2x = _max.x(), p2y, p2z;

  if (sb == sa) {
    p1y = _min.y(); p2y = _max.y();
  } else {
    p1y = _max.y(); p2y = _min.y();
  }

  if (sc == sa) {
    p1z = _min.z(); p2z = _max.z();
  } else {
    p1z = _max.z(); p2z = _min.z();
  }

  double dot1 = n.x()*p1x + n.y()*p1y + n.z()*p1z + d ;
  double dot2 = n.x()*p2x + n.y()*p2y + n.z()*p2z + d ;
  bool sd1 = dot1 >= 0;
  bool sd2 = dot2 >= 0;
  return (sd1 != sd2);
}

// looking for silhouettes recursively
void OcTree::findSilhouettes(const cvec3d& n, double d, Silhouettes *sil) {
  if (_isLeaf) {
    for (int i = 0; i < _edges.size(); i++)
      if (!_edges[i]->checked()) {
	_edges[i]->check();
	bool b1 = n.dot(_edges[i]->v(0)) + d > 0;
	bool b2 = n.dot(_edges[i]->v(1)) + d > 0;
	if (b1 != b2) {
	  sil->insertEdge(_edges[i]);
	}
      }
  } else {
    for (int i = 0; i < 8; i++)
      if (_child[i])
	if (_child[i]->touchPlane(n, d))
	  _child[i]->findSilhouettes(n, d, sil);
  }
}

// build octree
OcTree* OcTree::buildTree(TMesh *mesh) {
  OcTree* tree = new OcTree[8];

  // clear boundaries
  boundary.clear();

  // loop over all triangles, insert all leaves to the tree
  TMesh::FaceIterType ti = mesh->faceBegin();
  for (; ti != mesh->faceEnd(); ++ti) {
    if ((*ti)->isLeaf()) {
      for(EnoType e = 1; e <= (*ti)->noVtx(); ++e) {
	DualEdge::project(tree, *ti, e);
      }
    }
  }

  for (int j = 0; j < 8; j++) tree[j].split();

  return tree;
}
