// -*- mode: c++ -*-
// $Id: silhouette.h,v 1.14 2000/07/23 19:54:17 jianbo Exp $
// $Source: /tools/cvs-repos/illustrate/src/silhouette.h,v $

#ifndef __SILHOUETTE_H__
#define __SILHOUETTE_H__

#ifdef _WIN32
#include <set>
#else
#include <hash_set>
#endif

#include "camera.h"
#include "grid3d.h"
#include "dualmesh.h"

class SilEdge;

// intersection of silhouette edges, only information of the occluded edge
// are stored
class Intersection {
public:
  Intersection() { };
  Intersection(SilEdge *se, SilEdge *se1, double par, Camera *vc);
  ~Intersection() { _e = 0; };

  // edge near the camera
  SilEdge* e() { return _e; };
  // intersection position
  double p() { return _p; };
  // whether v(0) is occluded
  bool headOccluded() { return _headOccluded; };

  // new and delete operators
  void* operator new(size_t);
  void  operator delete(void *) { };
  // clear memory
  static void clear();

private:
  // blocking edge
  SilEdge *_e;
  // intersect parameter
  double _p;
  bool _headOccluded;

  // memory management
  static int _curr, _curr_chunk;
  static vector<Intersection*> _mem;
};

// silhouette edge (internal)
class SilEdge {
public:
  SilEdge() { };
  SilEdge( DualEdge *e, Camera *vc );
  SilEdge( BoundaryEdge *e, Camera *vc );
  ~SilEdge();

  // boundary ?
  bool isBoundary() const { return _be != 0; };

  // already put on chain ?
  bool onChain() const { return _onChain; };
  void setOnChain(bool b) { _onChain = b; };

  // put into cluster
  int cluster() const { return _cluster; };
  void clusterize( int c );

  // front or back
  bool facing() const
    { if (_de) return _de->facing(); else return true; };
  // vertices, triangles and edge numbers
  Vertex* v(int i) { return _v[i]; };
  Tri* tri(int i)
    { if (_de) return _de->tri(i); else return _be->tri(); };
  EnoType eno(int i) const
    { if (_de) return _de->eno(i); else return _be->eno(); };
  // position in image plane
  const cvec3d& iv(int i) const { return _iv[i]; };

  // intersections where this edge is farther from camera
  int nInter() { return _intersection.size(); };
  Intersection* intersection(int i) { return _intersection[i]; };
  // intersection positions
  cvec3d inter_pos(int i)
    { return _iv[0] + _intersection[i]->p()*(_iv[1] - _iv[0]); };

  // intersect with another edge
  void intersect(SilEdge *e, Camera *vc);
  // edges checked for intersection
  bool intersected(SilEdge *e) {
    int isz = _intersected.size();
    for (int i = 0; i < isz; i++)
      if (_intersected[i] == e) return true;
    return false;
  };
  // sort intersection from one vertex to the other
  void sortIntersection(bool fromhead);

  // how many triangles in the triangle ring are between this edge
  // and the camera
  int blockingFaces(Vertex *center, const cvec3d & viewpos);

  // new and delete operators
  void* operator new(size_t);
  void  operator delete(void *) { };
  // clear memory
  static void clear();

private:
  bool _onChain;
  int _cluster;

  // edge in dual mesh or boundary
  DualEdge *_de;
  BoundaryEdge *_be;
  // vertices
  Vertex *_v[2];
  // position in the image plane
  cvec3d _iv[2];

  // intersections where this edge is further from camera
  vector<Intersection*> _intersection;
  // intersected edges, used for efficiency
  vector<SilEdge*> _intersected;

  // memory management
  static int _curr, _curr_chunk;
  static vector<SilEdge*> _mem;
};

// silhouette chain
class Chain {
public:
  Chain() : _qi(-1) { };
  ~Chain() { _edges.clear(); _points.clear(); };

  // head and tail parameters
  double head() const { return _head; };
  double tail() const { return _tail; };

  // chain size (number of edges)
  int size() const { return _edges.size(); };
  // edges on chain
  SilEdge* edge(int i) { return _edges[i]; };
  // points on chain
  vector<cvec3d>& points() { return _points; };

  // add an edge to the chain
  void setHead(SilEdge *e, double p);
  void insert(SilEdge *e, bool fromhead);
  void setTail(SilEdge *e, double p);

  // quantity invisibility
  int QI() { return _qi; };
  void setQI(int q) { _qi = q; };
  void modifyQI(int q) { _qi += q; };
  bool visible() { return _qi == 0; };

  // new and delete operators
  void* operator new(size_t);
  void  operator delete(void *) { };
  // clear memory
  static void clear();

private:
  // head and tail parameters
  double _head, _tail;

  vector<SilEdge*> _edges;
  vector<cvec3d> _points;

  int _qi;

  // memory management
  static int _curr, _curr_chunk;
  static vector<Chain*> _mem;
};

// 2D grid for intersection computation
class Grid2D {
public:
  Grid2D(int s = 10);
  ~Grid2D();

  // grid size
  int size() const { return _size; };
  // cell length and width
  double xstep() const { return _xstep; };
  double ystep() const { return _ystep; };
  // bounding box, should be 2D vector, use cvec3d
  const cvec3d minp() const { return _min; };
  const cvec3d maxp() const { return _max; };

  vector<SilEdge*>* operator[] (int i) { return _grid[i]; };

  void placeIntoGrid(vector<SilEdge*>& edges);

  // clear all edges in grid
  void clear();

private:
  int _size;
  double _xstep, _ystep;
  cvec3d _min, _max;
  vector<SilEdge*> ** _grid;
};

// silhouettes
class Silhouettes {
public:
  Silhouettes() : _grid(0), _meshgrid(0), _camera(0) { };
  Silhouettes(Grid3D *g, Camera *c) : _grid(0), _meshgrid(g), _camera(c) { };
  ~Silhouettes();

  vector<SilEdge*>& edges() { return _edges; };
  vector<Chain*>& chains() { return _chains; };
  Grid2D* grid() { return _grid; };

  // put edge
  void insertEdge(DualEdge *e);
  void insertEdge(BoundaryEdge *e);

  // clear all chains
  void clear();
  // create grid
  void createGrid(int size);
  // put edges into clusters
  void clusterize();
  // find intersections of edges
  void findIntersections();
  // create chains from edges
  void createChains();

private:
  // silhouette edges and chains
  vector<SilEdge*> _edges;
  vector<Chain*> _chains;
  // 2D grid for silhouettes
  Grid2D *_grid;
  // 3D grid for mesh
  Grid3D *_meshgrid;
  // camera
  Camera *_camera;

  // triangle pair used to record inserted edges
  typedef pair<Tri*, Tri*> tripair;

#ifdef _WIN32
  set<tripair> siltripairs;
#else

  struct tripair_equal {
    bool operator() (const tripair& p1, const tripair& p2) const {
      return ((p1.first == p2.first && p1.second == p2.second) ||
	      (p1.first == p2.second && p1.second == p2.first));
    }
  };

  struct tripair_hash {
    size_t operator() (const tripair& p) const {
      return (p.first > p.second)?
	9973 * (long) p.first + (long) p.second :
	9973 * (long) p.second + (long) p.first;
    }
  };

  hash_set<tripair, tripair_hash, tripair_equal> siltripairs;
#endif
};

#endif /* __SILHOUETTE_H__ */
