// -*- Mode: c++ -*-
// $Id: arcball.cpp,v 1.3 2000/07/19 19:08:16 jianbo Exp $
// $Source: /tools/cvs-repos/illustrate/sub/arcball.cpp,v $

/* Subdivide V2.0
   Copyright (C) 2000 Henning Biermann, Denis Zorin, NYU

This file is part of Subdivide.

Subdivide is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Subdivide is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Subdivide; see the file COPYING.  If not, write to the Free
Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */


#include "compat.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include "arcball.h"
#include <algorithm>

 
#define LG_NSEGS 6
#define NSEGS (1<<LG_NSEGS)

#define RIMCOLOR()    glColor3f(1.0f, 1.0f, 1.0f)
#define FARCOLOR()    glColor3f(0.75f, 0.5f, 0.125f)
#define NEARCOLOR()   glColor3f(1.0f, 1.0f, 0.25f)
#define DRAGCOLOR()   glColor3f(0.5f, 1.0f, 1.0f)
#define RESCOLOR()    glColor3f(0.75f, 0.125f, 0.125f)


// ******************** auxilary math. functions ****************************

// Convert window coordinates to sphere coordinates. 


cvec3d ArcBall::RayOnSphere(const cvec3d& ray_start, const cvec3d& ray_dir) {
  cvec3d ballpoint;
  double ray_proj = (-center+ray_start).dot(ray_dir);

  cvec3d ray_perp = (-center+ray_start) - ray_proj*ray_dir;
  if( ray_perp.l2() > radius ) {
    ballpoint = ray_perp.dir();
  } else { 
     double start_center_distsq = (center-ray_start).dot();
	 double Discr = max(0.0, double(ray_proj*ray_proj - 
		 start_center_distsq + radius*radius) );
     double t1 = -ray_proj + sqrt(Discr);
	 double t2 = -ray_proj - sqrt(Discr);
	 double t = (fabs(t1) > fabs(t2))? t2:t1;
     ballpoint = ray_start + t*ray_dir - center;
  }
  return (ballpoint);
}


/* Force sphere point to be perpendicular to axis. */
static cvec3d ConstrainToAxis(const cvec3d& loose, const cvec3d& axis)
{
  cvec3d onPlane;
  double norm;
  onPlane = loose - axis*axis.dot( loose);
  norm = onPlane.l2();
  if (norm > 0.0) {
    return onPlane.dir();
  } /* else drop through */
  if (axis.z() == 1.0) {
    onPlane = cvec3d(1.0, 0.0, 0.0);
  } else {
    onPlane = (cvec3d(-axis.y(), axis.x(), 0.0)).dir();
  }
  return (onPlane);
}

/* Find the index of nearest arc of axis set. */
static int NearestConstraintAxis(cvec3d loose, cvec3d *axes, int nAxes)
{
  cvec3d onPlane;
  double max, dp;
  int i, nearest;
  max = -1.0; nearest = 0;
  for (i=0; i<nAxes; i++) {
    onPlane = ConstrainToAxis(loose, axes[i]);
    dp = onPlane.dot(loose);

    if (dp > max) {
      max = dp; nearest = i;
    }
  }
  return (nearest);
}

// ******************** drawing helpers ********************

// Draw an arc defined by its ends.
static void DrawAnyArc(const cvec3d& vFrom, const cvec3d& vTo)
{
    int i;
    cvec3d pts[NSEGS+1];
    double dp;
    pts[0] = vFrom;
    pts[1] = pts[NSEGS] = vTo;
    for (i=0; i<LG_NSEGS; i++) 
      pts[1] = (pts[0] + pts[1]).dir();

    dp = 2.0* pts[0].dot(pts[1]);

    for (i=2; i<NSEGS; i++) {
      pts[i] = pts[i-1]*dp - pts[i-2];
    }

    glBegin(GL_LINE_STRIP);
    for (i=0; i<=NSEGS; i++)
	glVertex3dv(pts[i]);
    glEnd();
}

// Draw a big circle in the plane perp. to the vector
static void DrawFullArc(const cvec3d& n)
{
    cvec3d p(0), m(0);
    p.z() = 0.0;
    if ((n.x() == 0.0) && (n.y() == 0.0)) {
      p.x() = 0.0; p.y() = 1.0;
    } else {
      p.x() = n.y(); 
      p.y() = -n.x();
    }
    p = p.dir(); 


    m = p.cross(n);

    glBegin(GL_LINES);
	glVertex3d(0.0,0.0,0.0);
	glVertex3dv(2*n );
	glEnd();

    DrawAnyArc(p, m);
    DrawAnyArc(m, -p);
    DrawAnyArc(-p, -m);
    DrawAnyArc(-m,p);
 }



// ******************** ArcBall memeber functions ********************

void ArcBall::Init()
{
  static cvec3d unitVecs[3] = { cvec3d(1.0, 0.0, 0.0), 
                                cvec3d(0.0, 1.0, 0.0), 
                                cvec3d(0.0, 0.0, 1.0) };
  radius = 1.0;
  center = cvec3d(0.0,0.0,0.0);
  
  
  rayStartDown = rayStartNow = cvec3d(0.0,0.0,0.0);
  rayDirDown = rayDirNow = cvec3d(0.0,0.0,1.0);

  
  qDown = qNow = Quat(0.0,0.0,0.0,1.0);
  mNow.setIdentity();
  mDown.setIdentity();
   
  combined.setIdentity();

  showResult = dragging = GL_FALSE;
  axisSet = NoAxes;
  sets[WorldAxes][X] = unitVecs[X];     
  sets[WorldAxes][Y] = unitVecs[Y];     
  sets[WorldAxes][Z] = unitVecs[Z];     
  setSizes[WorldAxes] = 3;
  sets[BodyAxes][X] = mDown.col(X); 
  sets[BodyAxes][Y] = mDown.col(Y); 
  sets[BodyAxes][Z] = mDown.col(Z); 
  setSizes[BodyAxes] = 3;

  outlineColor  = cvec3f(1.0f, 1.0f, 1.0f);
  inactiveColor = cvec3f(0.75f, 0.5f, 0.125f);
  activeColor   = cvec3f(1.0f, 1.0f, 0.25f);
  dragColor     = cvec3f(0.5f, 1.0f, 1.0f);
  resultColor   = cvec3f(0.75f, 0.125f, 0.125f);

  Update();
}


void ArcBall::Update()
{
    int setSize = setSizes[axisSet];
    cvec3d* axset = sets[axisSet];

    vFrom = RayOnSphere(rayStartDown,rayDirDown).dir();
    vTo   = RayOnSphere(rayStartNow,rayDirNow).dir();

    if (dragging) {
      if (axisSet != NoAxes) {
	vFrom = ConstrainToAxis(vFrom, axset[axisIndex]);
	vTo   = ConstrainToAxis(  vTo, axset[axisIndex]);
      }
      qDrag = Quat(vFrom, vTo);
      qNow = qDrag* qDown;
    } else {
      if (axisSet != NoAxes) {
	axisIndex = NearestConstraintAxis(vTo, axset, setSize);
      }
    }
    qDown.ballPoints(vrFrom, vrTo);
    SetRotation( HMatrix( qNow ) );


    if( X != axisIndex ) sets[BodyAxes][X] = mNow.col(X); 
    if( Y != axisIndex ) sets[BodyAxes][Y] = mNow.col(Y); 
    if( Z != axisIndex ) sets[BodyAxes][Z] = mNow.col(Z); 

}


void ArcBall::BeginDrag() { 
 dragging = GL_TRUE; 
 rayStartDown = rayStartNow; 
 rayDirDown = rayDirNow;
}

void ArcBall::EndDrag() {  
  dragging = GL_FALSE;
  qDown = qNow; mDown = mNow;
  sets[BodyAxes][X] = mDown.col(X); 
  sets[BodyAxes][Y] = mDown.col(Y); 
  sets[BodyAxes][Z] = mDown.col(Z); 
}

void ArcBall::DrawCompleteBall(const cvec3d& camerapos) {
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glTranslatef(center.x(),center.y(), center.z());  

  DrawOuterRing(camerapos);
  glScalef(radius, radius, radius);
  
  DrawResultArc();
  DrawConstraints(); 
  DrawDragArc();
  
  glPopMatrix();

}

// Draw the controller with all its arcs.
void ArcBall::Draw(const cvec3d& camerapos) {
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glTranslatef(center.x(),center.y(), center.z());

  if (axisSet != NoAxes)  DrawOuterRing(camerapos);

  glScalef(radius, radius, radius);
  
  DrawResultArc();
  DrawConstraints();
  DrawDragArc();

  glPopMatrix();

}


// Draw all constraint arcs.
void ArcBall::DrawConstraints() {

    if (axisSet==NoAxes) return;

    cvec3d* axset = sets[axisSet];

    int axisI, setSize = setSizes[axisSet];

    for (axisI=0; axisI < setSize; axisI++) {
      if (axisIndex != axisI) {
        //        if (dragging) continue;
        glColor3fv(inactiveColor);
	  } else  glColor3fv(activeColor);
      DrawFullArc(axset[axisI]);
    }
}

void ArcBall::DrawOuterRing(const cvec3d& camerapos) {

  glColor3fv(outlineColor);
  cvec3d cameradir = (camerapos - center).dir();
  double d = (camerapos - center).l2();


  glPushMatrix();
  cvec3d displ = cameradir*(radius*radius/d);
  glTranslatef(displ.x(), displ.y(), displ.z() );
  double newradius = radius*sqrt(1.0 - radius*radius/(d*d) );

  glScalef( newradius, newradius, newradius );
  DrawFullArc(cameradir);
  glPopMatrix();

}

// Draw "rubber band" arc during dragging.
void ArcBall::DrawDragArc() {
  if (dragging) { 
    glColor3fv(dragColor);
    DrawAnyArc(vFrom, vTo);
  }
}

// Draw arc for result of all drags.
void ArcBall::DrawResultArc()
{
    glColor3fv(resultColor);
    if (showResult) DrawAnyArc(vrFrom, vrTo);
}







