// -*- Mode: c++ -*-
// $Id: trirule.cpp,v 1.3 2000/07/19 19:08:16 jianbo Exp $
// $Source: /tools/cvs-repos/illustrate/sub/trirule.cpp,v $

/* Subdivide V2.0
   Copyright (C) 2000 Henning Biermann, Denis Zorin, NYU

This file is part of Subdivide.

Subdivide is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Subdivide is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Subdivide; see the file COPYING.  If not, write to the Free
Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "compat.h"
#include <math.h>
#include "trirule.h"

static void computeBoundaryParams(double k,
				  double& thetak,
				  double& delta,
				  double& gamma,
				  double& a,
				  double& b,
				  double& theta,
				  double& sigma1, 
				  double& sigma2, 
				  double& sigma3,
				  double th = 0);
  
double Coef::dot(const Coef& coef) const {
  assert(coef.edgeC.size() == edgeC.size());
  double s = centerC * coef.centerC;
  for(uint i = 0; i < edgeC.size(); ++i)
    s += edgeC[i] * coef.edgeC[i];
  return s;
}

double Coef::dotOne() const {
  double s = centerC;
  for(uint i = 0; i < edgeC.size(); ++i)
    s += edgeC[i];
  return s;
}

//----------------------------------------------------------------------------

InteriorTriRule::InteriorTriRule(int k) : k(k) {

  computeL0();
  computeL1();
  computeL2();
  computeX1();
  computeX2();
  computeLambda1();
  computeLambda2();
  computeSubCoef();
  computeEdgeSubCoef();
}

void InteriorTriRule::computeEdgeSubCoef() {
  _edgeSub.c[0] = 3.0/8.0; 
  _edgeSub.c[1] = 1.0/8.0;
  _edgeSub.c[2] = 3.0/8.0; 
  _edgeSub.c[3] = 1.0/8.0;
}

void InteriorTriRule::computeSubCoef() {

  double beta;

  if(k == 3) 
    beta = 3.0/16.0;
  else 
    beta = 3.0/8.0/(double)k;  
    
  _sub.edgeC = vector<double>(k);
  _sub.centerC = 1 - (double) k * beta;
  
  for (int i = 0; i < k; ++i)
    _sub.edgeC[i] = beta;
    
}
  
void InteriorTriRule::computeL0() {
  _l0.centerC = 0.5; 
  _l0.edgeC = vector<double>(k);
  if(k == 3) {
    _l0.centerC = 2/5.0;
    for(int i = 0; i < k; ++i)
      _l0.edgeC[i] = 1/5.0;
  } else {
    for (int i = 0; i < k; ++i) 
      _l0.edgeC[i] = 0.5/(double)k;
  }
}

void InteriorTriRule::computeL1() {
  _l1.centerC = 0;
  _l1.edgeC = vector<double>(k);
  for (int i = 0; i < k; ++i) 
    _l1.edgeC[i] = 2/(double) k * sin(2.0*M_PI*i/(double)k);
}

void InteriorTriRule::computeL2() {
  _l2.centerC = 0.0;
  _l2.edgeC = vector<double>(k);
  for (int i = 0; i < k; ++i)
    _l2.edgeC[i] = 2/(double) k * cos(2.0*M_PI*i/(double)k);
}

void InteriorTriRule::computeX1() {
  _x1.centerC = 0.0;
  _x1.edgeC = vector<double>(k);
  for (int i = 0; i < k; ++i) 
    _x1.edgeC[i] = sin(2.0*M_PI*i/(double)k);
}

void InteriorTriRule::computeX2() {
  _x2.centerC = 0.0;
  _x2.edgeC = vector<double>(k);
  for (int i = 0; i < k; ++i) 
    _x2.edgeC[i] = cos(2.0*M_PI*i/(double)k);
}

void InteriorTriRule::computeLambda1() 
{ _lambda1 = 3/8.0 + 1/4.0 * cos(2*M_PI/(double)k); }

void InteriorTriRule::computeLambda2()
{  _lambda2 = 3/8.0 + 1/4.0 * cos(2*M_PI/(double)k); }

  
//--------------------------------------------------------------------------------

CreaseTriRule::CreaseTriRule(int k) : k(k) {

  computeL0();
  computeL1();
  computeL2();
  computeX1();
  computeX2();
  computeLambda1();
  computeLambda2();
  computeSubCoef();
  computeEdgeSubCoef();
}

void CreaseTriRule::computeEdgeSubCoef() {
  double gamma = 1.0/2.0 - 1.0/4.0 * cos(M_PI/double(k));
  _edgeSub.c[0] = 3.0/4.0-gamma; 
  _edgeSub.c[1] = 1.0/8.0;
  _edgeSub.c[2] = gamma;
  _edgeSub.c[3] = 1.0/8.0; 
}

void CreaseTriRule::computeSubCoef() {
  _sub.edgeC = vector<double>(k+1);
  _sub.centerC = 6/8.0;
  _sub.edgeC[0] = 1/8.0;
  _sub.edgeC[k] = 1/8.0;
  for (int i = 1; i < k; ++i)
    _sub.edgeC[i] = 0.0;
}

void CreaseTriRule::computeL0() {
  _l0.edgeC = vector<double>(k+1);
  _l0.centerC = 2/3.0;
  _l0.edgeC[0] = 1/6.0;
  _l0.edgeC[k] = 1/6.0;
  for (int i = 1; i < k; ++i)
    _l0.edgeC[i] = 0.0;
}

void CreaseTriRule::computeL1() {
  _l1.edgeC = vector<double>(k+1);
  if (k == 1) {
    _l1.centerC = -1.0;
    _l1.edgeC[0] = 0.5;
    _l1.edgeC[1] = 0.5;
  } else {
    double thetak, delta, gamma, theta, a, b, sigma1, sigma2, sigma3;
    computeBoundaryParams(k, thetak, delta,gamma,
			  a, b, theta,
			  sigma1, sigma2,sigma3);
      
    _l1.edgeC[0] = -(2.0/(double)k)*(( 0.5*a + 1.0/6.0)*sigma1 + 0.5*b*sigma3);
    _l1.centerC  = -(2.0/(double)k)*(( 2.0/3.0 - a)*sigma1 - b*sigma3);
    _l1.edgeC[k] = _l1.edgeC[0];
      
    for (int i = 1; i < k; i++ )
      _l1.edgeC[i] = (2.0/(double)k)*sin(thetak*(double)i);
  }
}

void CreaseTriRule::computeL2() {
  _l2.edgeC = vector<double>(k+1);
  _l2.centerC = 0.0;
  _l2.edgeC[0] = 0.5;
  _l2.edgeC[k] = -0.5;
  for (int i = 1; i < k; ++i)
    _l2.edgeC[i] = 0.0;
}

void CreaseTriRule::computeX1() {
  _x1.edgeC = vector<double>(k+1);
    
  if (k == 1) {
    _x1.centerC =  double(-1/3.0);
    _x1.edgeC[0] = double(2.0/3.0);
    _x1.edgeC[1] = double(2.0/3.0);
  } else {
    _x1.centerC = 0.0;
    for (int i = 0; i < k+1; ++i) 
      _x1.edgeC[i] = sin(i * M_PI/(double) k);
  }
}

void CreaseTriRule::computeX2() {
  _x2.edgeC = vector<double>(k+1);
  if (k == 1) {
    _x2.centerC = 0;
    _x2.edgeC[0] = 1.0;
    _x2.edgeC[1] = -1.0;
  } else {
    _x2.centerC = 0.0;
    for (int i = 0; i < k+1; ++i)
      _x2.edgeC[i] = cos(i * M_PI / (double) k);
  }
}

void CreaseTriRule::computeLambda1() {
  if (k == 1) 
    _lambda1 = 0.25;
  else
    _lambda1 = 0.5;
}

void CreaseTriRule::computeLambda2() {
  _lambda2 = 0.5;
}

//--------------------------------------------------------------------------------

ConcaveTriRule::ConcaveTriRule(int k, double theta) : k(k), theta(theta) {

  computeL0();
  computeL1();
  computeL2();
  computeX1();
  computeX2();
  computeLambda1();
  computeLambda2();
  computeSubCoef();
  computeEdgeSubCoef();
}

void ConcaveTriRule::computeEdgeSubCoef() {
  double gamma = 1.0/2.0 - 1.0/4.0 * cos(theta/double(k));
  _edgeSub.c[0] = 3.0/4.0-gamma; 
  _edgeSub.c[1] = 1.0/8.0;
  _edgeSub.c[2] = gamma;
  _edgeSub.c[3] = 1.0/8.0; 
}

void ConcaveTriRule::computeSubCoef() {
  _sub.edgeC = vector<double>(k+1);
  _sub.centerC = 1.0;
  for(int i = 0; i < k+1; ++i)
    _sub.edgeC[i] = 0.0;
}

void ConcaveTriRule::computeL0() {
  _l0.edgeC = vector<double>(k+1);
  _l0.centerC = 1.0;
  for (int i = 0; i < k+1; ++i) 
    _l0.edgeC[k] = 0.0;
}

void ConcaveTriRule::computeL1() {
  _l1.edgeC = vector<double>(k+1);
  _l1.centerC = -1.0;
  _l1.edgeC[0] = 1.0;
  for (int i = 1; i < k+1; ++i) 
    _l1.edgeC[k] = 0.0;
}

void ConcaveTriRule::computeL2() {
  _l2.edgeC = vector<double>(k+1);
  _l2.centerC = -1.0;
  _l2.edgeC[k] = 1.0;
  for (int i = 0; i < k; ++i) 
    _l2.edgeC[i] = 0.0;
}
  
void ConcaveTriRule::computeX1() {
  double thetak = theta / (double) k;
  _x1.edgeC = vector<double>(k+1);
  _x1.centerC = 0.0;
  for (int i = 0; i < k+1; ++i) 
    _x1.edgeC[i] = sin((k-i)*thetak) / sin(theta);
}

void ConcaveTriRule::computeX2() {
  double thetak = theta / (double) k;
  _x2.edgeC = vector<double>(k+1);
  _x2.centerC = 0.0;
  for (int i = 0; i < k+1; ++i) 
    _x2.edgeC[i] = sin(i * thetak) / sin(theta);
}

void ConcaveTriRule::computeLambda1() {
  _lambda1 = 0.5;
}

void ConcaveTriRule::computeLambda2() {
  _lambda2 = 0.5;
}

//--------------------------------------------------------------------------------

ConvexTriRule::ConvexTriRule(int k, double theta) : k(k), theta(theta) {

  computeL0();
  computeL1();
  computeL2();
  computeX1();
  computeX2();
  computeLambda1();
  computeLambda2();
  computeSubCoef();
  computeEdgeSubCoef();
}

void ConvexTriRule::computeEdgeSubCoef() {
  double gamma = 1.0/2.0 - 1.0/4.0 * cos(theta/double(k));
  _edgeSub.c[0] = 3.0/4.0-gamma; 
  _edgeSub.c[1] = 1.0/8.0;
  _edgeSub.c[2] = gamma;
  _edgeSub.c[3] = 1.0/8.0; 
}

void ConvexTriRule::computeSubCoef() {
  _sub.edgeC = vector<double>(k+1);
  _sub.centerC = 1.0;
  for(int i = 0; i < k+1; ++i)
    _sub.edgeC[i] = 0.0;
}
  
void ConvexTriRule::computeL0() {
  _l0.edgeC = vector<double>(k+1);
  _l0.centerC = 1.0;
  for (int i = 0; i < k+1; ++i) 
    _l0.edgeC[k] = 0.0;
}

void ConvexTriRule::computeL1() {
  _l1.edgeC = vector<double>(k+1);
  _l1.centerC =  -1.0;
  _l1.edgeC[k] =  1.0;
  for (int i = 0; i < k; ++i)
    _l1.edgeC[i] =  0.0;
}
  
void ConvexTriRule::computeL2() {
  _l2.edgeC = vector<double>(k+1);
  _l2.centerC =  -1.0;
  _l2.edgeC[0] =  1.0;
  for (int i = 1; i < k+1; ++i)
    _l2.edgeC[i] = 0.0;
}

void ConvexTriRule::computeX1() {
  _x1.edgeC = vector<double>(k+1);
  _x1.centerC = 0.0;
  for (int i = 0; i < k+1; ++i) 
    _x1.edgeC[i] = sin(i*theta/(double)k)/sin(theta);
}

void ConvexTriRule::computeX2() {
  _x2.edgeC = vector<double>(k+1);
  _x2.centerC = 0.0;
  for (int i = 0; i < k+1; ++i)
    _x2.edgeC[i] = sin((k-i)*theta/(double)k)/sin(theta);
}

void ConvexTriRule::computeLambda1() 
{ _lambda1 =  0.5; }

void ConvexTriRule::computeLambda2() 
{ _lambda2 = 0.5; }
  

void computeBoundaryParams(double k,
			   double& thetak,
			   double& delta,
			   double& gamma,
			   double& a,
			   double& b,
			   double& theta,
			   double& sigma1, 
			   double& sigma2, 
			   double& sigma3,
			   double th) {
    
    
  if (th == 0) 
    thetak = M_PI / (double) k;
  else
    thetak = th / (double) k;
    
  delta = 1.0/8.0;
  gamma = 0.5 - 0.25*cos(thetak);
  theta = acos( 0.5*(0.25 - gamma)/delta);
    
  a = (1.0 - 2.0*delta - gamma)/(3.0*(2.0*delta + gamma - 0.25));
  b = (2.0/3.0 - a)/cos(k*theta/2.0);
    
  sigma1 = sin(thetak)/(1.0 - cos(thetak));
  sigma2 = (cos(2.0*thetak) + 1.0 + 2.0*cos(thetak))/
    (sin(thetak) + sin(2.0*thetak));
  sigma3 = cos( k*theta/2.0)*sin(thetak)/(cos(theta) - cos(thetak));
    
}

